@@ -23,6 +23,9 @@ along with APAV.  If not, see <http://www.gnu.org/licenses/>.
 from sys import getsizeof
 from functools import lru_cache
 import os
+import struct
+from enum import Enum
+from warnings import warn
 
 import numpy as n
 import numpy.linalg as la
@@ -159,7 +162,7 @@ class Roi:
         Get the x/y/z dimensions of the dataset
         """
         # return tuple(i[1] - i[0] for i in self.xyz_extents)
-        return n.diff(self.xyz_extents)
+        return n.diff(self.xyz_extents).ravel()
 
     @property
     def mass_extents(self) -> Tuple[float, float]:
@@ -219,6 +222,243 @@ class Roi:
         """
         return "tof" in self.misc.keys()
 
+    @classmethod
+    def from_apt(cls, filepath: str, verbose: bool = False):
+        """
+        Read the contents of an apt file into a Roi container
+
+        :param filepath: Path to apt file
+        :param verbose: Print the structure of the apt file as it is read (for debug purposes)
+        """
+
+        validate.file_exists(filepath)
+
+        class RelType(Enum):
+            REL_UNKNOWN = 0
+            ONE_TO_ONE = 1
+            INDEXED = 2,
+            UNRELATED = 3
+            ONE_TO_MANY = 4
+
+        class RecordType(Enum):
+            RT_UNKNOWN = 0
+            FIXED_SIZE = 1
+            VARIABLE_SIZE = 2
+            VARIABLE_INDEXED = 3
+
+        class RecordDataType(Enum):
+            DT_UNKNOWN = 0
+            INT = 1
+            UINT = 2
+            FLOAT = 3
+            CHARSTRING = 4
+            OTHER = 5
+
+        class Dtype(Enum):
+            int32 = 4
+            int64 = 8
+            char = 1
+            wchar_t = 2
+            filetime = 8
+
+        def record_dtype2numpy_dtype(rec_dtype: RecordDataType, size: int):
+            """
+            Map a section's record data type to its equivalent numpy dtype
+            """
+            if rec_dtype in (RecordDataType.UINT, RecordDataType.CHARSTRING):
+                raise ValueError("Cannot map to UINT or CHARSTRING")
+
+            int_map = {
+                8: n.int8,
+                16: n.int16,
+                32: n.int32,
+                64: n.int64
+            }
+
+            float_map = {
+                32: n.float32,
+                64: n.float64
+            }
+
+            if rec_dtype == RecordDataType.INT:
+                return int_map[size]
+            elif rec_dtype == RecordDataType.FLOAT:
+                return float_map[size]
+            else:
+                raise ValueError(f"Unexpected record data type {rec_dtype}")
+
+        # Maps the apt format data type to str format needed for struct.unpack
+        dtype2fmt = {
+            Dtype.int32: "i",
+            Dtype.int64: "q",
+            Dtype.char: "c",
+            Dtype.filetime: "Q",
+            Dtype.wchar_t: "c"
+        }
+
+        # Maps the apt format data type to python data type
+        dtype2constr = {
+            Dtype.int32: int,
+            Dtype.int64: int,
+            Dtype.char: lambda x: x.decode("utf-8"),
+            Dtype.wchar_t: lambda x: x.decode("utf-16"),
+            Dtype.filetime: int
+        }
+
+        with open(filepath, "rb") as dat:
+            def read_chunk(dtype: Dtype,
+                           count: int = 1,
+                           start: Union[None, int] = None) -> Union[Tuple[Any], Any]:
+                if isinstance(start, int):
+                    dat.seek(start)
+
+                fmt = dtype2fmt[dtype]*count
+                constructor = dtype2constr[dtype]
+                dtype_size = dtype.value
+
+                if dtype in (Dtype.wchar_t, Dtype.char):
+                    return constructor(dat.read(dtype_size*count)).replace("\x00", "")
+                else:
+                    retn = struct.unpack("<" + fmt, dat.read(dtype_size * count))
+
+                if len(retn) == 1:
+                    return constructor(retn[0])
+                else:
+                    return tuple(constructor(i) for i in retn)
+
+            cSignature = read_chunk(Dtype.char, 4)
+
+            # Read the APT file header --------------------------------------------------------------------------------
+            iHeaderSize = read_chunk(Dtype.int32)
+            iHeaderVersion = read_chunk(Dtype.int32)
+            wcFileName = read_chunk(Dtype.wchar_t, 256)
+            ftCreationTime = read_chunk(Dtype.filetime)
+            llIonCount = read_chunk(Dtype.int64)
+
+            if verbose:
+                print(f"\nReading header of {filepath}")
+                print(f"\tcSignature: " + cSignature)
+                print(f"\tiHeaderSize: {iHeaderSize}")
+                print(f"\tiHeaderVersion: {iHeaderVersion}")
+                print(f"\twcFileName: {wcFileName}")
+                print(f"\tftCreationTime: {ftCreationTime}")
+                print(f"\t11IonCount: {llIonCount}")
+
+            # Read the APT sections ----------------------------------------------------------------------------
+            section_start = iHeaderSize
+            section_data = {}
+
+            while True:
+                sec_sig = read_chunk(Dtype.char, 4, section_start)
+
+                if sec_sig == '':
+                    # EOF reached
+                    break
+
+                # Flag used to not include a section in the Roi when a configuration
+                # situation is not implemented or handled
+                skip_sec = False
+
+                sec_header_size = read_chunk(Dtype.int32)
+                sec_header_ver = read_chunk(Dtype.int32)
+                sec_type = read_chunk(Dtype.wchar_t, 32)
+                sec_ver = read_chunk(Dtype.int32)
+
+                sec_rel_type = RelType(read_chunk(Dtype.int32))
+                is_one_to_one = sec_rel_type == RelType.ONE_TO_ONE
+                if not is_one_to_one:
+                    warn(f"APAV does not handle REL_TYPE != ONE_TO_ONE, section \"{sec_type}\" will be ignored")
+                    skip_sec = True
+
+                sec_rec_type = RecordType(read_chunk(Dtype.int32))
+                is_fixed_size = sec_rec_type == RecordType.FIXED_SIZE
+                if not is_fixed_size:
+                    warn(f"APAV does not handle RECORD_TYPE != FIXED_SIZE, section \"{sec_type}\" will be ignored")
+                    skip_sec = True
+
+                sec_rec_dtype = RecordDataType(read_chunk(Dtype.int32))
+                if sec_rec_dtype in (RecordDataType.DT_UNKNOWN, RecordDataType.OTHER, RecordDataType.CHARSTRING):
+                    warn(f"APAV does not handle RECORD_TYPE == {sec_rec_dtype}, section \"{sec_type}\" will be ignored")
+                    skip_sec = True
+
+                sec_dtype_size = read_chunk(Dtype.int32)
+                sec_rec_size = read_chunk(Dtype.int32)
+                sec_data_unit = read_chunk(Dtype.wchar_t, 16)
+                sec_rec_count = read_chunk(Dtype.int64)
+                sec_byte_count = read_chunk(Dtype.int64)
+
+                if verbose:
+                    print("\nReading new section")
+                    print(f"\tSection header sig: {sec_sig}")
+                    print(f"\tSection header size: {sec_header_size}")
+                    print(f"\tSection header version: {sec_header_ver}")
+                    print(f"\tSection type: {sec_type}")
+                    print(f"\tSection version: {sec_ver}")
+                    print(f"\tSection relative type: {sec_rel_type}")
+                    print(f"\tSection record type: {sec_rec_type}")
+                    print(f"\tSection record data type: {sec_rec_dtype}")
+                    print(f"\tSection data type size (bits): {sec_dtype_size}")
+                    print(f"\tSection record size: {sec_rec_size}")
+                    print(f"\tSection data type unit: {sec_data_unit}")
+                    print(f"\tSection record count: {sec_rec_count}")
+                    print(f"\tSection byte count: {sec_byte_count}")
+
+                if not skip_sec:
+                    columns = int(sec_rec_size/(sec_dtype_size/8))
+                    records = int(sec_rec_count)
+                    count = records*columns
+                    in_data = n.fromfile(filepath,
+                                         record_dtype2numpy_dtype(sec_rec_dtype, sec_dtype_size),
+                                         count,
+                                         offset=section_start+sec_header_size)
+                    if columns > 1:
+                        section_data[sec_type] = in_data.reshape(records, columns)
+                    else:
+                        section_data[sec_type] = in_data
+
+                section_start = section_start + sec_byte_count + sec_header_size
+
+        has_mass_data = "Mass" in section_data.keys()
+        has_pos_data = "Position" in section_data.keys()
+
+        # Map some APT section names to names that APAV expects, otherwise the provided name is retained
+        name_map = {
+            "Multiplicity": "ipp",
+            "Time of Flight": "tof",
+            "XDet_mm": "det_x",
+            "YDet_mm": "det_y",
+            "Voltage": "dc_voltage",
+            "Pulse Voltage": "pulse_voltage"
+        }
+
+        # Require mass and position data, clean up some sections, and account for possible duplicate sections (i.e.
+        # XDet_mm + YDet_mm combined with Detector Coordinates
+        if not has_mass_data:
+            raise AttributeError("APT file must have include a mass section")
+        elif not has_pos_data:
+            raise AttributeError("APT file must have include a position section")
+
+        mass = section_data.pop("Mass")
+        pos = section_data.pop("Position")
+
+        # There are 2 difference ways that detector space coordinates can be included in an apt file, as a single
+        # section containing both x/y or the x and y in separate sections. Only when the separate x/y sections are not
+        # present we will load the combined x/y data (which we separate into different x and y arrays).
+        if "Detector Coordinates" in section_data.keys():
+            temp = section_data.pop("Detector Coordinates")
+            if "XDet_mm" not in section_data.keys():
+                section_data["det_x"] = temp[:, 0]
+            if "YDet_mm" not in section_data.keys():
+                section_data["det_y"] = temp[:, 1]
+
+        roi = cls(pos, mass)
+        roi._filepath = filepath
+        for key, data in section_data.items():
+            name = key if key not in name_map.keys() else name_map[key]
+            roi.misc[name] = data
+
+        return roi
+
     @classmethod
     def from_pos(cls, filepath: str):
         """
@@ -686,14 +926,11 @@ class DummyRoiHistogram(Roi):
         :param x: the x values of the mass histogram
         :param y: the y values of the mass histogram
         """
-        super().__init__(n.array([[0, 0, 0]]), n.array([[0]]))
+        super().__init__(n.array([[0, 0, 0]]), n.array([0]))
         self.__histogram = (x, y)
 
     def mass_histogram(self, *args, **kwargs):
         """
         Override :meth:`Roi.mass_histogram()` to always return the specified mass histogram
-
-        :param args:
-        :param kwargs:
         """
         return self.__histogram
